@@ -1,5 +1,5 @@
 /* hdparm.c - Command line interface to get/set hard disk parameters */
-/*          - by Mark S. Lord (c) 1994 -- freely distributable */
+/*          - by Mark S. Lord (c) 1994-1996 -- freely distributable */
 
 #include <unistd.h>
 #include <string.h>
@@ -19,8 +19,14 @@
 #include <linux/fs.h>
 #include <linux/major.h>
 
-#ifndef HD1_MAJOR
-#define HD1_MAJOR 22	/* for second controller using atdisk2 patches */
+/*
+ * For kernels prior to 1.3.61, HDIO_SET_32BIT is HDIO_SET_CHIPSET
+ */
+#ifndef HDIO_SET_32BIT
+#ifdef HDIO_SET_CHIPSET
+#define	HDIO_SET_32BIT HDIO_SET_CHIPSET
+#define	HDIO_GET_32BIT HDIO_GET_CHIPSET
+#endif
 #endif
 
 #define TIMING_MB		16
@@ -31,18 +37,27 @@
 
 char *progname;
 static struct hd_driveid id;
-static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1;
+static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
 static int flagcount = 0, do_flush = 0, do_timings = 0, is_scsi_hd = 0;
 
 static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
 static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;     
 static unsigned long set_unmask   = 0, get_unmask   = 0, unmask   = 0;     
 static unsigned long set_mult     = 0, get_mult     = 0, mult     = 0;     
+#ifdef HDIO_SET_DMA
+static unsigned long set_dma      = 0, get_dma      = 0, dma      = 0;     
+#endif
+#ifdef HDIO_SET_NOWERR
+static unsigned long set_nowerr   = 0, get_nowerr   = 0, nowerr   = 0;
+#endif
 #ifdef HDIO_GET_KEEPSETTINGS
 static unsigned long set_keep     = 0, get_keep     = 0, keep     = 0;
 #endif
-#ifdef HDIO_SET_CHIPSET
-static unsigned long set_chipset  = 0, get_chipset  = 0, chipset  = 0;
+#ifdef HDIO_SET_32BIT
+static unsigned long set_io32bit  = 0, get_io32bit  = 0, io32bit  = 0;
+#endif
+#ifdef HDIO_SET_PIO_MODE
+static unsigned long set_piomode  = 0, noisy_piomode= 0, piomode  = 0;
 #endif
 #ifdef HDIO_DRIVE_CMD
 static unsigned long set_dkeep    = 0, get_dkeep    = 0, dkeep    = 0;
@@ -53,6 +68,9 @@ static unsigned long set_prefetch = 0, get_prefetch = 0, prefetch = 0;
 static unsigned long set_wcache   = 0, get_wcache   = 0, wcache   = 0;
 static unsigned long set_seagate  = 0, get_seagate  = 0;
 #endif
+#ifdef HDIO_SET_32BIT
+static int get_IDentity = 0;
+#endif
 
 char *cfg_str[] =
 {	"", " HardSect", " SoftSect", " NotMFM", " HdSw>15uSec", " SpinMotCtl",
@@ -74,7 +92,7 @@ static void dump_identity (struct hd_driveid *id)
 {
 	int i;
 
-	printf("\n Model=%.40s, FwRev=%.20s, SerialNo=%.8s", 
+	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.8s", 
 		id->model, id->fw_rev, id->serial_no);
 	printf("\n Config={");
 	for (i=0; i<=15; i++) {
@@ -86,11 +104,11 @@ static void dump_identity (struct hd_driveid *id)
 		id->cyls, id->heads, id->sectors,
 		id->track_bytes, id->sector_bytes, id->ecc_bytes);
 	dmpstr (" BuffType=",id->buf_type,BuffType,3);
-	printf(", BuffSize=%dKB, MaxMultSect=%d", id->buf_size/2, id->max_multsect);
+	printf(", BuffSize=%dkB, MaxMultSect=%d", id->buf_size/2, id->max_multsect);
 	if (id->max_multsect) {
 		printf(", MultSect=");
 		if (!(id->multsect_valid&1))
-			printf("???");
+			printf("?%d?", id->multsect);
 		else if (id->multsect)
 			printf("%d", id->multsect);
 		else
@@ -132,8 +150,10 @@ static void dump_identity (struct hd_driveid *id)
 			if (id->dma_mword & 1)		printf("mword0 ");
 			if (id->dma_mword & 0x200)	printf("*");
 			if (id->dma_mword & 2)		printf("mword1 ");
-			if (id->dma_mword & 0xfc00)	printf("*");
-			if (id->dma_mword & 0xfc)	printf("mword? ");
+			if (id->dma_mword & 0x400)	printf("*");
+			if (id->dma_mword & 4)		printf("mword2 ");
+			if (id->dma_mword & 0xf800)	printf("*");
+			if (id->dma_mword & 0xf8)	printf("mword? ");
 		}
 	}
 	if ((id->capability&8) || (id->field_valid&2)) {
@@ -285,12 +305,14 @@ void time_device (int fd)
 
 	if (elapsed <= (correction * 2))
 		printf("Hmm.. suspicious results: probably not enough free memory for a proper test.\n");
+#if 0  /* the "estimate" is just plain wrong for many systems.. */
 	else {
 		printf(" Estimating raw driver speed: ");
 		elapsed -= correction;
 		printf("%2d MB in %5.2f seconds =%5.2f MB/sec\n",
 			TIMING_MB, elapsed, TIMING_MB / elapsed);
 	}
+#endif
 quit:
 	if (-1 == shmdt(buf))
 		perror ("could not detach sharedmem buf");
@@ -405,9 +427,22 @@ void process_dev (char *devname)
 		exit(errno);
 	}
 
-	is_scsi_hd = (major(stat_buf.st_rdev) == SCSI_DISK_MAJOR);
-	if (major(stat_buf.st_rdev) != HD_MAJOR
-	 && major(stat_buf.st_rdev) != HD1_MAJOR && !is_scsi_hd) {
+	if ((major(stat_buf.st_rdev) == SCSI_DISK_MAJOR) 
+#ifdef MD_MAJOR
+		|| (major(stat_buf.st_rdev) == MD_MAJOR)
+#endif
+							) 
+		is_scsi_hd = 1;
+	else
+		if (major(stat_buf.st_rdev) != IDE0_MAJOR
+		 && major(stat_buf.st_rdev) != IDE1_MAJOR
+#ifdef IDE2_MAJOR
+		 && major(stat_buf.st_rdev) != IDE2_MAJOR
+#endif
+#ifdef IDE3_MAJOR
+		 && major(stat_buf.st_rdev) != IDE3_MAJOR
+#endif
+	                                                      ) {
 		fprintf(stderr,"%s is not a hard disk.\n",devname);
 		exit(EINVAL);
 	}
@@ -418,7 +453,7 @@ void process_dev (char *devname)
 		exit(errno);
 	}
 
-	if (get_readahead)
+	if (!quiet)
 		printf("\n%s:\n", devname);
 	if (set_readahead) {
 		if (get_readahead)
@@ -426,13 +461,26 @@ void process_dev (char *devname)
 		if (ioctl(fd, BLKRASET, readahead)) 
 			perror(" BLKRASET failed");
 	}
-#ifdef HDIO_SET_CHIPSET
-	if (set_chipset) {
+#ifdef HDIO_SET_PIO_MODE
+	if (set_piomode) {
 		no_scsi();
-		if (get_chipset)
-			printf(" setting chipset to %ld\n", chipset);
-		if (ioctl(fd, HDIO_SET_CHIPSET, chipset)) 
-			perror(" HDIO_SET_CHIPSET failed");
+		if (noisy_piomode) {
+			if (piomode == 255)
+				printf(" attempting to auto-tune PIO mode\n");
+			else
+				printf(" attempting to set PIO mode to %ld\n", piomode);
+		}
+		if (ioctl(fd, HDIO_SET_PIO_MODE, piomode)) 
+			perror(" HDIO_SET_PIO_MODE failed");
+	}
+#endif
+#ifdef HDIO_SET_32BIT
+	if (set_io32bit) {
+		no_scsi();
+		if (get_io32bit)
+			printf(" setting 32-bit I/O support flag to %ld\n", io32bit);
+		if (ioctl(fd, HDIO_SET_32BIT, io32bit)) 
+			perror(" HDIO_SET_32BIT failed");
 	}
 #endif
 	if (set_mult) {
@@ -462,6 +510,28 @@ void process_dev (char *devname)
 		if (ioctl(fd, HDIO_SET_UNMASKINTR, unmask))
 			perror(" HDIO_SET_UNMASKINTR failed");
 	}
+#ifdef HDIO_SET_DMA
+	if (set_dma) {
+		no_scsi();
+		if (get_dma) {
+			printf(" setting using_dma to %ld", dma);
+			on_off(dma);
+		}
+		if (ioctl(fd, HDIO_SET_DMA, dma))
+			perror(" HDIO_SET_DMA failed");
+	}
+#endif
+#ifdef HDIO_SET_NOWERR
+	if (set_nowerr) {
+		no_scsi();
+		if (get_nowerr) {
+			printf(" setting nowerr to %ld", nowerr);
+			on_off(nowerr);
+		}
+		if (ioctl(fd, HDIO_SET_NOWERR, nowerr))
+			perror(" HDIO_SET_NOWERR failed");
+	}
+#endif
 #ifdef HDIO_GET_KEEPSETTINGS
 	if (set_keep) {
 		no_scsi();
@@ -558,14 +628,6 @@ void process_dev (char *devname)
 	if (!flagcount)
 		verbose = 1;
 
-	if (verbose || get_readahead) {
-		if (ioctl(fd, BLKRAGET, &parm))
-			perror(" BLKRAGET failed");
-		else {
-			printf(" readahead    = %2ld", parm);
-			on_off(parm);
-		}
-	}
 	if ((verbose && !is_scsi_hd) || get_mult || get_identity) {
 		no_scsi();
 		multcount = -1;
@@ -576,21 +638,23 @@ void process_dev (char *devname)
 			on_off(multcount);
 		}
 	}
-	if (verbose || get_readonly) {
-		if (ioctl(fd, BLKROGET, &parm))
-			perror(" BLKROGET failed");
-		else {
-			printf(" readonly     = %2ld", parm);
-			on_off(parm);
-		}
-	}
-#ifdef HDIO_GET_CHIPSET
-	if (get_chipset) {
+#ifdef HDIO_GET_32BIT
+	if ((verbose && !is_scsi_hd) || get_io32bit) {
 		no_scsi();
-		if (ioctl(fd, HDIO_GET_CHIPSET, &parm))
-			perror(" HDIO_GET_CHIPSET failed");
+		if (ioctl(fd, HDIO_GET_32BIT, &parm))
+			perror(" HDIO_GET_32BIT failed");
 		else {
-			printf(" chipset   =%3ld\n", parm);
+			printf(" I/O support  =%3ld (", parm);
+			switch (parm) {
+				case 0:	printf("default ");
+				case 2: printf("16-bit)\n");
+					break;
+				case 1:	printf("32-bit)\n");
+					break; 
+				case 3:	printf("32-bit w/sync)\n");
+					break; 
+				default:printf("???)\n");
+			}
 		}
 	}
 #endif
@@ -603,15 +667,17 @@ void process_dev (char *devname)
 			on_off(parm);
 		}
 	}
-	if (verbose || get_geom) {
-		static struct hd_geometry g;
-		if (ioctl(fd, BLKGETSIZE, &parm))
-			perror(" BLKGETSIZE failed");
-		else if (ioctl(fd, HDIO_GETGEO, &g))
-			perror(" HDIO_GETGEO failed");
-		else	printf(" geometry     = %d/%d/%d, sectors = %ld, start = %ld\n",
-				g.cylinders, g.heads, g.sectors, parm, g.start);
+#ifdef HDIO_GET_DMA
+	if ((verbose && !is_scsi_hd) || get_dma) {
+		no_scsi();
+		if (ioctl(fd, HDIO_GET_DMA, &parm))
+			perror(" HDIO_GET_DMA failed");
+		else {
+			printf(" using_dma    = %2ld", parm);
+			on_off(parm);
+		}
 	}
+#endif
 #ifdef HDIO_GET_KEEPSETTINGS
 	if ((verbose && !is_scsi_hd) || get_keep) {
 		no_scsi();
@@ -624,6 +690,42 @@ void process_dev (char *devname)
 	}
 #endif /* HDIO_GET_KEEPSETTINGS */
 
+#ifdef HDIO_SET_NOWERR
+	if (verbose || get_nowerr) {
+		no_scsi();
+		if (ioctl(fd, HDIO_GET_NOWERR, &parm))
+			perror(" HDIO_GET_NOWERR failed");
+		else {
+			printf(" nowerr       = %2ld", parm);
+			on_off(parm);
+		}
+	}
+#endif
+	if (verbose || get_readonly) {
+		if (ioctl(fd, BLKROGET, &parm))
+			perror(" BLKROGET failed");
+		else {
+			printf(" readonly     = %2ld", parm);
+			on_off(parm);
+		}
+	}
+	if ((verbose && !is_scsi_hd) || get_readahead) {
+		if (ioctl(fd, BLKRAGET, &parm))
+			perror(" BLKRAGET failed");
+		else {
+			printf(" readahead    = %2ld", parm);
+			on_off(parm);
+		}
+	}
+	if (verbose || get_geom) {
+		static struct hd_geometry g;
+		if (ioctl(fd, BLKGETSIZE, &parm))
+			perror(" BLKGETSIZE failed");
+		else if (ioctl(fd, HDIO_GETGEO, &g))
+			perror(" HDIO_GETGEO failed");
+		else	printf(" geometry     = %d/%d/%d, sectors = %ld, start = %ld\n",
+				g.cylinders, g.heads, g.sectors, parm, g.start);
+	}
 	if (get_identity) {
 		no_scsi();
 		if (!(ioctl(fd, HDIO_GET_IDENTITY, &id))) {
@@ -638,6 +740,16 @@ void process_dev (char *devname)
 		else
 			perror(" HDIO_GET_IDENTITY failed");
 	}
+#ifdef HDIO_SET_32BIT
+	if (get_IDentity) {
+		unsigned char args[4+512] = {WIN_IDENTIFY,0,0,1,};
+		no_scsi();
+		if (!(ioctl(fd, HDIO_DRIVE_CMD, &args)))
+			dump_identity((struct hd_driveid *)&args[4]);
+		else
+			perror(" HDIO_DRIVE_CMD failed");
+	}
+#endif
 	if (do_timings)
 		time_device (fd);
 	if (do_flush)
@@ -647,44 +759,56 @@ void process_dev (char *devname)
 
 void usage_error (void)
 {
-	fprintf(stderr,"%s - get/set hard disk parameters - version 2.3\n\n", progname);
+	fprintf(stderr,"\n%s - get/set hard disk parameters - version 2.9\n\n", progname);
 	fprintf(stderr,"Usage:  %s  [options] [device] ..\n\n", progname);
-	fprintf(stderr,"Options:\n");
-	fprintf(stderr," -a    get/set fs readahead\n");
+	fprintf(stderr,"Options:\n"
+	" -a    get/set fs readahead\n"
 #ifdef HDIO_DRIVE_CMD
-	fprintf(stderr," -A  * set drive read-lookahead flag (0/1)\n");
+	" -A  * set drive read-lookahead flag (0/1)\n"
 #endif
-#ifdef HDIO_SET_CHIPSET
-	fprintf(stderr," -c    get/set IDE chipset/interface type\n");
+#ifdef HDIO_SET_32BIT
+	" -c  * get/set IDE 32-bit IO setting\n"
+#endif
+#ifdef HDIO_SET_DMA
+	" -d  * get/set using_dma flag\n"
+#endif
+	" -f    flush buffer cache for device on exit\n"
+	" -g    display drive geometry\n"
+	" -h    display terse usage information\n"
+	" -i  * display drive identification\n"
+#ifdef HDIO_SET_32BIT
+	" -I  * read drive identification directly from drive\n"
 #endif
-	fprintf(stderr," -f    flush buffer cache for device on exit\n");
-	fprintf(stderr," -g    display drive geometry\n");
-	fprintf(stderr," -h    display terse usage information\n");
-	fprintf(stderr," -i  * display drive identification\n");
 #ifdef HDIO_GET_KEEPSETTINGS
-	fprintf(stderr," -k  * get/set keep_settings_over_reset flag (0/1)\n");
+	" -k  * get/set keep_settings_over_reset flag (0/1)\n"
 #endif
 #ifdef HDIO_DRIVE_CMD
-	fprintf(stderr," -K  * set drive keep_features_over_reset flag (0/1)\n");
+	" -K  * set drive keep_features_over_reset flag (0/1)\n"
+#endif
+	" -m  * get/set multiple sector count\n"
+#ifdef HDIO_SET_NOWERR 
+	" -n  * get/set ignore-write-errors flag (0/1)\n"
+#endif
+#ifdef HDIO_SET_PIO_MODE
+	" -p  * set PIO mode on IDE interface chipset (0,1,2,3,4,5)\n"
 #endif
-	fprintf(stderr," -m  * get/set multiple sector count\n");
 #ifdef HDIO_DRIVE_CMD
-	fprintf(stderr," -P  * set drive prefetch count\n");
+	" -P  * set drive prefetch count\n"
 #endif
-	fprintf(stderr," -q    change next setting quietly\n");
-	fprintf(stderr," -r    get/set readonly flag\n");
+	" -q    change next setting quietly\n"
+	" -r    get/set readonly flag (DANGEROUS to set)\n"
 #ifdef HDIO_DRIVE_CMD
-	fprintf(stderr," -S  * set standby (spindown) timeout\n");
+	" -S  * set standby (spindown) timeout\n"
 #endif
-	fprintf(stderr," -t    perform device read timings\n");
-	fprintf(stderr," -u  * get/set unmaskirq flag (0/1) (DANGEROUS)\n");
-	fprintf(stderr," -v    (default) same as -agkmru (-agr for SCSI)\n");
+	" -t    perform device read timings\n"
+	" -u  * get/set unmaskirq flag (0/1) (DANGEROUS)\n"
+	" -v    default; same as -acdgkmnru (-gr for SCSI)\n"
 #ifdef HDIO_DRIVE_CMD
-	fprintf(stderr," -W  * set drive write-caching flag (0/1) (DANGEROUS)\n");
-	fprintf(stderr," -X  * set IDE xfer mode (DANGEROUS)\n");
-	fprintf(stderr," -Z  * disable Seagate auto-powersaving mode\n");
+	" -W  * set drive write-caching flag (0/1) (DANGEROUS)\n"
+	" -X  * set IDE xfer mode (DANGEROUS)\n"
+	" -Z  * disable Seagate auto-powersaving mode\n"
 #endif
-	fprintf(stderr,"        * = (E)IDE drives only\n\n");
+	"        * = (E)IDE drives only\n\n");
 	exit(1);
 }
 
@@ -719,6 +843,11 @@ int main(int argc, char **argv)
 					case 'v':
 						verbose = 1;
 						break;
+#ifdef HDIO_SET_32BIT
+					case 'I':
+						get_IDentity = 1;
+						break;
+#endif
 					case 'i':
 						get_identity = 1;
 						break;
@@ -729,6 +858,7 @@ int main(int argc, char **argv)
 						do_flush = 1;
 						break;
 					case 'q':
+						quiet = 1;
 						noisy = 0;
 						break;
 					case 'u':
@@ -741,6 +871,43 @@ int main(int argc, char **argv)
 							unmask = *p++ - '0';
 						}
 						break;
+#ifdef HDIO_SET_DMA
+					case 'd':
+						get_dma = noisy;
+						noisy = 1;
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
+						if (*p == '0' || *p == '1') {
+							set_dma = 1;
+							dma = *p++ - '0';
+						}
+						break;
+#endif
+#ifdef HDIO_SET_NOWERR
+					case 'n':
+						get_nowerr = noisy;
+						noisy = 1;
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
+						if (*p == '0' || *p == '1') {
+							set_nowerr = 1;
+							nowerr = *p++ - '0';
+						}
+						break;
+#endif
+#ifdef HDIO_SET_PIO_MODE
+					case 'p':
+						noisy_piomode = noisy;
+						noisy = 1;
+						if (!*p && argc && isdigit(**argv))
+							p = *argv++, --argc;
+						if (*p >= '0' && *p <= '5')
+							piomode = *p++ - '0';
+						else
+							piomode = 255;  /* auto-tune */
+						set_piomode = 1;
+						break;
+#endif
 					case 'r':
 						get_readonly = noisy;
 						noisy = 1;
@@ -756,13 +923,13 @@ int main(int argc, char **argv)
 						noisy = 1;
 						GET_NUMBER(set_mult,mult);
 						break;
-#ifdef HDIO_SET_CHIPSET
+#ifdef HDIO_SET_32BIT
 					case 'c':
-						get_chipset = noisy;
+						get_io32bit = noisy;
 						noisy = 1;
-						GET_NUMBER(set_chipset,chipset);
+						GET_NUMBER(set_io32bit,io32bit);
 						break;
-#endif /* HDIO_SET_CHIPSET */
+#endif /* HDIO_SET_32BIT */
 #ifdef HDIO_DRIVE_CMD
 					case 'S':
 						get_standby = noisy;
